/*-------------------------------------------------------------
 *  Copyright (c) 2010-2015, MSS ComBlock. All rights reserved.
 *
 * 	Filename: main.c
 *  Last modified: 12/12/15
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 4.03
 *  Inheritance: comblock1400 Atmel code
 *  Description: main program. Includes main loop for handling commands on a polling basis.
 *
 *  Dependencies/Pre-requisite: N/A
 *
 *  3/2/14 AZ
 *  Added Get Multiple Registers GMR and GMT commands for faster GUI
 *  Added Get Multiple Status registers GMS for faster GUI
 *
 *  Rev 4.03 12/12/15 AZ
 *  Added usb connection/disconnection in main, depending on the presence of the USB VBUS
 -------------------------------------------------------------*/
#define __MAIN_C__

#include "main.h"
#include "uart.h"
#include "utils.h"
#include "nandFlash.h"
#include "iap.h"
#include "fpga.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
volatile unsigned long SysTickCnt; /* SysTick Counter                    */
unsigned long cpuFreq; // CPU frequency in Hz.
volatile unsigned long timer[NTIMERS];
// timers, expressed in multiple of the system tick (10ms). 0 means expired.
// volatile because the Systick ISR will modify the values
// timer[0] = NAND flash timer
// timer[1] = forced delay to save the config into flash
// timer[2] = inactivity timer
volatile struct string rxLine; // rx command line
volatile struct string txLine; // tx command line
// separate rx and tx lines (unlike previous Atmel code). Allows concurrent tx and rx and interrupts.
char address[4]; // ComBlock address, assigned dynamically (string of 3 characters + \0 end of string)
long commandPortNo; // port through which this board receives commands: 1 = top (USB ARM), 2 = right, 3 = bottom (USB FPGA), 4 = left (ComBlock rule)
unsigned char forwardMask; // Mask output ports. bit0 = top port1, bit1 = right port 2, etc. '1' means the port is blocked, '0' means the port is enabled for output
struct config_hw configHw; // hardware-related configuration.
struct config_sw configSw; // software/FPGA-related configuration volatile
char regv[NREGS]; // control registers, volatile copy of configSw.reg[]
int active_sw_config; // current active configuration (generally the same as default_sw_config, but it could be temporary different until the new FPGA config is loaded)
int usb_softconnect;	// 0 for no vbus/no connect, 1 for vbus/connected
/***************************************************************************/
/*  main                                                                   */
/***************************************************************************/
int main(void) {
	// local variables
	unsigned char a;
	unsigned int i, j, c, reg_no, s78_decode, s789_decode, retval, packet_no,
	rx_byte_count, l, block_no, phyBlockNo, page_no, startCol, nreg_no, start_reg_no;
	int sn[4];
	int sel;
	struct config_sw cs; // work area for a software configuration (don't store anything of value here)
	rx_byte_count = 0;

	// initial ARM processor configuration (clock frequency, system tick, etc)
	initConfig();
	// reset FPGA configuration (just in case it was interfering with the ARM - NAND Flash shared data bus
rst:

	fpga_reset();
	// initialize UARTs 0 and 3
	uart_init();
	// initialize NAND Flash
	nandFlash_init();
	// When the flash sector CONFIG_SECTOR is blank (all FFs), initialize it to 'all zeros'.
	initialize_configs();
	// load last-saved configuration from non-volatile (flash) memory
	read_hw_config(&configHw);
	// assess JP1 once at power-up/reset
	assess_JP1();
	// the active config is the last FPGA config, unless the JP1 jumper is in position 2-3
	if(get_JP1val() == 23){
		// recovery mode. points to personality index 1 (generally protected at shipping)
		active_sw_config = 0;
	} else {
		// default configuration
		active_sw_config = configHw.default_sw_config;
	}
	// Semantics: we make a difference between the active_default_sw_config (currently loaded and running) and
	// default_sw_config which can be changed by the user prior to FPGA reboot.
	read_sw_config(active_sw_config, &configSw);
	memcpy(&regv, configSw.reg, sizeof(regv)); // initialize the temporary control registers with the previously saved ones.

	// configure the FPGA with the active software configuration
	// skip the initial FPGA configuration if JP1 is set to position 2-3 or if the software configuration is undefined
	if ((get_JP1val() != 23) && fpga_config_exists(configSw.fpga_start_address)) {
		// configure FPGA
		fpga_config(configSw.fpga_start_address);
	}

	// ComBlock address at power up is "000"
	snprintf(&address[0],sizeof(address), "000");

	// configure the FPGA with the 250 registers stored in non-volatile memory
	// skip registers 250,251,....,254 & 255, they are used to rx/tx data from FPGA
	for(i = 0; i < 250; i++){
		fpga_setByte(i, configSw.reg[i]);
	}

	// main loop -------------------------------------------------------------------------------
	while (1) {
		// USB softconnect management based on USB VBUS presence
		USB_connect_ctrl();

		// is it time to save the current active configuration?
		// timer[1] limits how frequently the configuration can be saved,
		// timer[2] indicates persistent inactivity.
		if (((configHw.save_request_pending) || (configSw.save_request_pending))
				&& (timer[2] == 0) && (timer[1] == 0)) {
			save_configuration(-1, &cs); // save the hardware + active software configurations. no alternate configuration.
		}

		// service UARTs (polling)
		uart_rxpoll();
		uart_txpoll();
		// service USB via FPGA (once the FPGA is configured)
		if (fpga_check_done()) {
			USB_FPGA_rxpoll();
		}

		// did we receive any message?
		// get a complete message from @ to \n (non-blocking)
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// When done processing the received message in rxLine, move the rptr = wptr to allow subsequent messages to be received.

			// re-arm inactivity timer because of new message just received.
			// We need to know when there is inactivity so that we can devote 120ms to update the configuration in flash.
			timer[2] = TIMER2;

			/*
			 // echo message (debugging)
			 memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
			 txLine.wptr = rxLine.wptr;
			 rxLine.rptr = rxLine.wptr;	// release rx message
			 sendLine(4);
			 continue;
			 */

			// Dispatch the message: local processing, ignore or forward?
			if (dispatchLine()) {
				// rx message is for us: address at the start of the rx message matches (exactly or 111 broadcast address)
				delay_ns(25000); // delay 25uS before responding (otherwise slower Atmel-based ComBlocks may not be ready for a reply)

				// Remember which port is used to communicate with ComBlock control center
				// Criteria: received complete command message destined to this particular ComBlock.
				commandPortNo = rxLine.portNo;

				// decode 2-digit decimal reg_no at position 7-8 in the received string. Used by most commands.
				s78_decode = rxLine.buf[7] - '0';
				s78_decode = s78_decode * 10 + rxLine.buf[8] - '0';

				// decode 3-digit decimal sector number at position 7-9 in the received string (common to many commands)
				s789_decode = s78_decode * 10 + rxLine.buf[9] - '0';
				block_no = s789_decode - 1; // translate ComBlock range 001-999 to NAND flash block number range starting at 0.


				// command handler
				if (!memcmp((const char *)&rxLine.buf[4], "SAC", 3)) {
					// SAC command - set address
					memcpy(&address[0], (const char *)&rxLine.buf[7], 3);

				} else if (!memcmp((const char *)&rxLine.buf[4], "MFW", 3)) {
					// MFW command - Message Forwarding
					// 0 = no forwarding, 1-4: forward to port 1-4
					// 9: forward to all ports (except of course the one where the message came from)
					if (rxLine.buf[7] == '0') { // stop all forwarding
						forwardMask = 0xFF;
					} else if (rxLine.buf[7] == '1') { // forward to top port
						forwardMask = 0xFE;
					} else if (rxLine.buf[7] == '2') { // forward to right port
						forwardMask = 0xFD;
					} else if (rxLine.buf[7] == '3') { // forward to bottom port
						forwardMask = 0xFB;
					} else if (rxLine.buf[7] == '4') { // forward to left port
						forwardMask = 0xF7;
					} else if (rxLine.buf[7] == '9') { // forward to all ports, but the command port itself
						if (commandPortNo == 1) {
							forwardMask = 0xF1;
						} else if (commandPortNo == 2) {
							forwardMask = 0xF2;
						} else if (commandPortNo == 3) {
							forwardMask = 0xF4;
						} else if (commandPortNo == 4) {
							forwardMask = 0xF8;
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "PNG", 3)) {
					// new PNG ping command. Echoes command back
					memcpy((char *)&(txLine.buf[0]), (const char *)&(rxLine.buf[0]),rxLine.wptr);
					txLine.buf[1]='9';
					txLine.buf[2]='9';
					txLine.buf[3]='9';
					txLine.wptr = rxLine.wptr;
					sendLine(commandPortNo);


				} else if (!memcmp((const char *)&rxLine.buf[4], "GMI", 3)) {
					// GMI - Get Module Identification command
					// Three cases:
					// (a) if the GMI command is followed by a non-zero FPGA configuration number,
					// then return the available information about that configuration,
					// (b) if the GMI command is followed by a zero configuration number,
					// then return the available information about the hardware module (FPGA platform)
					// (c) otherwise [if the GMI command is followed by CR or LF]
					// return the information about the active configuration
					// compute configuration number. comprises at exactly two digits.
					sel = 0;
 					if ((rxLine.buf[7] != '\r') && (rxLine.buf[7] != '\n')) {
						sel = s78_decode;
					} else {
						// no configuration selected, asking about active configuration
						sel = -1;
					}
					if (sel == 0) {
						// user is asking about hardware platform description
						// This is a COM-1X00 hardware platform with ARM rev x
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID" ARM_FW_REV "\r\n");
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo); // send response through the same port the query came in
					} else if ((sel <= NSWCONFIGS) && (sel >= -1)) {
						// user is asking about a software configuration
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID");
						if (sel == -1) {
							// no configuration selected, asking about active configuration
							memcpy(&cs, &configSw, sizeof(configSw));
							memcpy((char *)&(txLine.buf[7]), (const char *)&cs.mid[0], 4); // read 4-character mid
							txLine.buf[11] = fpga_getByte(0xfc); // option is read directly from FPGA itself
							txLine.buf[12] = fpga_getByte(0xfd); // revision is read directly from FPGA itself
						} else {
							// asking about a specific (not necessarily active) configuration
							sel--; // indices start at zero
							read_sw_config(sel, &cs); // read the software configuration of interest from flash
							memcpy((char *)&(txLine.buf[7]), (const char *)&cs.mid[0], 5); // read 5-character mid
							// non active configuration. We don't know the firmware rev no -> mask it
							txLine.buf[12] = ' ';
						}
						txLine.buf[13] = '0' + (unsigned char) (commandPortNo);
						// read configuration status: authorization/protection/etc  .
						// convert from binary to readable hex character
						// bit 0 = authorization
						// bit 1 = protection
						// bit 2 = FPGA DONE pin
						if (fpga_check_done())
							a = 0x04; // FPGA DONE pin is high
						else
							a = 0x00; // FPGA DONE pin is low

						// configuration is protected if boot section is protected (PROTECTION == 1)
						// or if protection flag is set for this configuration.
						if ((sel == 0) && (PROTECTION == 1)) {
							a = a | 0x02; // sector is protected
						} else if ((cs.protection) && (cs.mid[0] != 0xFF)) {
							a = a | 0x02; // module identification is defined and sector is protected
						}

						// TEMPORARY: authorized until hash code is in place
						a = a | 0x01; // always authorized (temporary)
						snprintf((char *)&txLine.buf[14], sizeof(txLine.buf) - 14, " %1X\r\n", a);
						// defensive code.. make sure there are no 0s. This could happen if the FPGA is not configured, or..
						for (i = 7; i <= 12; i++) {
							if (txLine.buf[i] == 0)
								txLine.buf[i] = '?';
						}
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo);
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SDC", 3)) {
					// SDC command - set default configuration
					// decode selected Configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if ((c != configHw.default_sw_config) && (c < NSWCONFIGS)) {
						// the default config changed.
						configHw.default_sw_config = c;
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						// immediate save (as this command is likely to be followed by an immediate reset)
						timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
						// Note: saving takes 120ms!! during which incoming messages are not processed
						save_configuration(-1, &cs);
						//NOTE: at this point, active and default configuration numbers are different.
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GDC", 3)) {
					// GDC command - Get Default Configuration command
					// prepare DCF response
					c = configHw.default_sw_config + 1; // display configuration number starting at 1 (more natural for the user)
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999DCF%02X\r\n", c);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SMI", 3)) {
					// SMI command - Set Module Identification

					// first check the length of the message.
					if ((rxLine.buf[12] == '\r') || (rxLine.buf[12] == '\n')) {
						// message is of the form @xxxSMI + MID + 1character option
						// thus 12 or 13 characters. use active configuration.
						// verify that the selected configuration is not protected unless
						// it is the first time this configuration is defined (all FF's)
						if ((!configSw.protection) || (configSw.mid[0] == 0xFF)) {
							memcpy(&configSw.mid[0], (const char *)&rxLine.buf[7], 5); // store MID + 1character option
							if (rxLine.buf[11] == '\r') // short message @xxxSMIYYYY\r\n
								configSw.mid[4] = ' '; // option cannot be '\r'
							configSw.mid[5] = 0; // don't know about FPGA firmware revision here
							configSw.protection = 0; // unprotected if first-time definition
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					} else if ((rxLine.buf[14] == '\r') || (rxLine.buf[14]
							== '\n')) {
						// message is of the form @xxxSMI+configuration number + MID + 1character option
						// thus 15 characters
						c = s78_decode - 1; // array indices start at 0, whereas configurations start at 1
						// make sure configuration pointer does not exceed array boundaries
						if (c < NSWCONFIGS) {
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							// verify that the selected configuration is not protected unless it is the first time this configuration is defined (all FF's)
							if ((!cs.protection) || (cs.mid[0] == 0xFF)) {
								memcpy(&cs.mid[0], (const char *)&rxLine.buf[9], 5); // store MID + 1character option
								cs.mid[5] = 0; // don't know about FPGA firmware revision here
								cs.protection = 0; // unprotected if first-time definition
								// immediate save this non-active configuration
								timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
								// Note: saving takes 120ms!! during which incoming messages are not processed
								save_configuration(c, &cs);
							}
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAI", 3)) {
					// SAI command - Set assembly identification
					// For security, allow only if not already programmed
					// first check if assembly id has already been programmed
					a = 0;
					/* temporarily disabled
					 for(i = 0; i < 4; i++){
					 if(aid[i] != 0)
					 a = 1;
					 }
					 */
					if (a == 0) {
						// if aid is undefined (all 0s), allow programming
						memcpy(&configHw.aid[0], (const char *)&rxLine.buf[7], 6);
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						timer[1] = TIMER1;
						// re-arm timer[1]: wait before saving the comblock configuration
						//into flash (to consolidate multiple successive changes into one sector programming operation)
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAS", 3)) {
					// SAS - Set Assembly Serial number,
					// For security, allow only if eeprom is not already programmed (i.e. 0000..)
					// first check if serial number has already been programmed
					a = 0;
					for (i = 0; i < sizeof(configHw.asn); i++) {
						if (configHw.asn[i] != 0)
							a = 1;
					}
					if (a == 0) {
						// if assembly serial number is null, allow programming
						memcpy(&configHw.asn[0], (const char *)&rxLine.buf[7],
								sizeof(configHw.asn));
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAI", 3)) {
					// GAI command - Get Assembly ID
					snprintf((char *)txLine.buf,sizeof(txLine.buf), "@999AID      \r\n"); // workaround problem with snprintf.
					memcpy((char *)&(txLine.buf[7]), (const char *)(configHw.aid), 6);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSN", 3)) {
					// GSN command - Get Serial Number
					// read unique LPC17xx device serial number
					read_sn(sn);
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MSN%10u %10u %10u %10u\r\n",
							sn[0], sn[1], sn[2], sn[3]); // unsigned decimal
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAS", 3)) {
					// GAS - Get Assembly Serial number command
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ASN%10s\r\n", configHw.asn);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "SRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "SRT", 3))) {
					// SRG - Set Register (persistent/non-volatile storage)
					// SRT - Set Register Temporary (i.e. volatile, lost when powering off)

					// compute register address. Address comprises at least two digits.
					c = s78_decode;
					// check whether destination register is 2 digit or 3 digit long
					if ((rxLine.buf[11] != '\r') && (rxLine.buf[11] != '\n')) {
						// three-digit register number
						c = c * 10 + rxLine.buf[9] - '0';
						// compute register value. Value starts at char 10
						a = hex2bin((char *)&rxLine.buf[10]);
					} else {
						// two-digit register number
						// compute register value. Value starts at char 9
						a = hex2bin((char *)&rxLine.buf[9]);
					}
					reg_no = c;
					// update the configuration register
					if (reg_no < sizeof(configSw.reg)) {
						regv[reg_no] = a;
						// write value to FPGA
						fpga_setByte(reg_no, a);
						// if SRG, also save in flash
						if (rxLine.buf[6] == 'G') {
							configSw.reg[reg_no] = a;
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSR", 3)) {
					// GSR command - Get Status Register
					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					//------> user to define status registers here <--------
					//------> user to define status registers here <--------
					// alternative status registers used for production tests (unused afterwards)
					a = 0;
					if (reg_no == 0)
						a = nfInfo.manufacturerID; // NAND flash manufacturer ID = 0x2C
					else if (reg_no == 1)
						a = nfInfo.deviceID; // NAND flash device ID = 0xD1
					else if (reg_no == 2)
						a = nfInfo.byte3; // NAND flash type = 0x95
					else if (reg_no == 3)
						a = nfInfo.nBadBlocks; // NAND number of bad blocks
					else if (reg_no == 4)
						a = power_status();
					else if (reg_no == 5) {
						if (fpga_check_done())
							a = 1;
						else
							a = 0;
					} else {
						a = fpga_getByte(reg_no);
					}
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GMS", 3)) {
					// new.
					// GMS command - Get multiple status registers (faster than asking for individual ones)
					// includes two parameters: number of registers to read (3 decimal digits) and start register (also 3 decimal digits)

					// number of registers to read
					c = rxLine.buf[7] - '0';
					c = c * 10 + rxLine.buf[8] - '0';
					c = c * 10 + rxLine.buf[9] - '0';
					nreg_no = c;
					// valid range is 0 - 128

					// starting at register number x
					c = rxLine.buf[10] - '0';
					c = c * 10 + rxLine.buf[11] - '0';
					c = c * 10 + rxLine.buf[12] - '0';
					start_reg_no = c;
					// valid range is 0 - 255

		 			// check length
		 			if((nreg_no > 128) || (start_reg_no > 255)){
		 				// return message would be too long or out of range
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
		 			} else {
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf),"@999MRV");	// Multiple Register Values
			 			for(i = 0; i< nreg_no; i++){
			 				reg_no = start_reg_no + i;
							a = 0;
							if (reg_no == 0)
								a = nfInfo.manufacturerID; // NAND flash manufacturer ID = 0x2C
							else if (reg_no == 1)
								a = nfInfo.deviceID; // NAND flash device ID = 0xD1
							else if (reg_no == 2)
								a = nfInfo.byte3; // NAND flash type = 0x95
							else if (reg_no == 3)
								a = nfInfo.nBadBlocks; // NAND number of bad blocks
							else if (reg_no == 4)
								a = power_status();
							else if (reg_no == 5) {
								if (fpga_check_done())
									a = 1;
								else
									a = 0;
							} else {
								a = fpga_getByte(reg_no);
							}
	              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",a);
			 			}
              			snprintf((char *)&txLine.buf[7 + 2*i],3,"\r\n");
		 			}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "GMR", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GMT", 3))) {
					// new. Get multiple registers (faster than asking for individual ones)
					// GMR command - get multiple registers (read from flash memory)
					// GMT command - get multiple registers temporary (read from volatile memory)

					// number of registers to read
					c = rxLine.buf[7] - '0';
					c = c * 10 + rxLine.buf[8] - '0';
					c = c * 10 + rxLine.buf[9] - '0';
					nreg_no = c;
					// valid range is 0 - 128

					// starting at register number x
					c = rxLine.buf[10] - '0';
					c = c * 10 + rxLine.buf[11] - '0';
					c = c * 10 + rxLine.buf[12] - '0';
					start_reg_no = c;
					// valid range is 0 - 255

		 			// check length
		 			if((nreg_no > 128) || (start_reg_no > 255)){
		 				// return message would be too long or out of range
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
		 			} else {
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf),"@999MRV");	// Multiple Register Values
			 			for(i = 0; i< nreg_no; i++){
							if (rxLine.buf[6] == 'R') {
								// read registers from flash memory
		              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",configSw.reg[start_reg_no+i]);
							} else {
								// read registers from volatile memory
		              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",regv[start_reg_no+i]);
							}
			 			}
              			snprintf((char *)&txLine.buf[7 + 2*i],3,"\r\n");
		 			}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "GRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GRT", 3))) {
					// GRG command - get register (read from flash memory)
					// GRGT command - get register temporary (read from volatile memory)

					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					if ((reg_no < sizeof(configSw.reg)) && (rxLine.buf[6]
							== 'G'))
						a = configSw.reg[reg_no]; // GRG, read non-volatile register
					else if ((reg_no < sizeof(regv)) && (rxLine.buf[6] == 'T')) // make sure we do not write past the buffer end
						a = regv[reg_no]; // GRT, read volatile register
					else
						// trying to read non-volatile registers past index NREG_V
						a = 0;

					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "EFS", 3)) {
					// EFS command - Erase File Sector
					// first check block number is valid
					if (block_no >= (NBLOCKS - nfInfo.nBadBlocks)) {
						retval = -1; // NAK
					} else {
						// make sure the block does not relate to a protected configuration
						c = block_to_config(block_no);
						read_sw_config(c, &cs); // read the software configuration of interest from flash
						if ((c == 0) && (PROTECTION == 1)) {
							retval = -1; // boot configuration is protected
						} else if (cs.protection) {
							retval = -1; // configuration is protected
						} else {
							retval = nandFlash_EFS(block_no);
						}
					}
					if (retval) {
						// error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK\r\n");
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);


				} else if (!memcmp((const char *)&rxLine.buf[4], "RMD", 3)) {
		 			// RMD - Read Memory Data command
		 			// write memory offset to FPGA registers 237 (LSB),238,239 (MSB)
					a = hex2bin((char *)&rxLine.buf[13]);
					fpga_setByte(237, a);	// write value to FPGA  REG237 LSB
					a = hex2bin((char *)&rxLine.buf[11]);
					fpga_setByte(238, a);	// write value to FPGA  REG238
					a = hex2bin((char *)&rxLine.buf[9]);
					fpga_setByte(239, a);	// write value to FPGA  REG239
		 			// byte in input_string[7:8] ignored. too large an MSB. Not supported in this FPGA

		 			// compute length of sequence to read
		 			// ignore 2 MSBs (not supported, as FPGA has only a few 4Kb dual-port RAM buffers
					l = 0;
					l |= hex2bin((char *)&rxLine.buf[19]) & 0x000000FF;
					l <<= 8;
					l |= hex2bin((char *)&rxLine.buf[21]) & 0x000000FF;

		 			// check length
		 			if(l > 128){
		 				// return message would be too long
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
		 			} else {
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf),"@999MDA");
			 			for(i = 0; i< l; i++){
			 				// read REG250 for data from FPGA DPRAM
			 				// autoincrement on
			 				a = fpga_getByte(0xFA);
	              			snprintf((char *)&txLine.buf[7 + 2*i],5,"%02X\r\n",a);
			 			}
		 			}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SCA", 3)) {
					// SCA - Set Configuration start Address command
					// read the selected configuration number for the specified start address
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						// decode 32-bit hex string, then save
						l = 0;
						l |= hex2bin((char *)&rxLine.buf[9]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[11]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[13]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[15]) & 0x000000FF;
						if (c == active_sw_config) {
							configSw.fpga_start_address = l;
							// save the active configuration
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							save_configuration(c, &configSw); // save the hardware + active software configurations
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							cs.fpga_start_address = l;
							// save this non-active configuration
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							// Note: saving takes 120ms!! during which incoming messages are not processed
							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GCA", 3)) {
					// GCA - Get Configuration start Address command
					// read the selected configuration number for the specified start address
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						if (c == active_sw_config) {
							// active configuration
							l = configSw.fpga_start_address;
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							l = cs.fpga_start_address;
						}
						// Virtual location (not accounting bad-block redirection to another physical address)
						if (l == 0xFFFFFFFF) {
							// undefined
							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS????????\r\n");
						} else {
							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS%08X\r\n", l);
						}
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo);
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "WFS", 3)) {
					// WFS - Write File Sector command

					// first check block number is valid
					if (block_no >= (NBLOCKS - nfInfo.nBadBlocks)) {
						retval = -1; // NAK
					} else {
						// make sure the block does not relate to a protected configuration
						c = block_to_config(block_no);
						read_sw_config(c, &cs); // read the software configuration of interest from flash
						if ((c == 0) && (PROTECTION == 1)) {
							retval = -1; // boot configuration is protected
						} else if (cs.protection) {
							retval = -1; // configuration is protected
						} else {
							// disassemble the packet sequence number and data field
							packet_no = 0;
							for (i = 0; i < 4; i++)
								packet_no = packet_no * 10 + rxLine.buf[i + 10]
										- '0';
							// convert the data field from hex ASCII to binary. Up to 128 bytes in the data field.
							l = (rxLine.wptr - 16) / 2;
							j = 0;
							for (i = 14; i < rxLine.wptr -2; i += 2) {
								rxLine.buf[j++] = hex2bin((char *)&rxLine.buf[i]); // save binary in place
							}

							// re-direct to a good block
							phyBlockNo = nfInfo.blockRedirect[block_no];
							if (packet_no == 1) {
								// first packet within a block. Reset the byte count.
								rx_byte_count = 0;
								page_no = 0;
								startCol = 0;
								// prepare the NAND flash for writing a block
								retval = nandFlash_page_program_start(phyBlockNo,
										page_no, startCol, l, (char *)rxLine.buf);
							} else {
								page_no = rx_byte_count / 2048;
								startCol = rx_byte_count % 2048;
								if (startCol == 0) {
									// prepare the NAND flash for writing the next page
									retval = nandFlash_page_program_start(
											phyBlockNo, page_no, startCol, l,
											(char *)rxLine.buf);
								} else {
									// continue writing to a previously opened page
									retval = nandFlash_program_contd(startCol,
											l, (char *)rxLine.buf);
								}
							}
							rx_byte_count += l;

							// program complete page?
							if (startCol + l >= 2048) {
								retval = nandFlash_page_program_end();
							} else if (packet_no == 9999) {
								retval = nandFlash_page_program_end();
							}
						}
					}

					if (retval) {
						// error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK%06d\r\n", rx_byte_count);
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GCS", 3)) {
					// GCS - Get CheckSum command
					retval = nandFlash_GCS(block_no, (unsigned long *) (&c));

					// first check block number is valid
					if (retval) {
						// block number out-of-range error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						// return the 16-bit checksum
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CKS%04X\r\n", c);
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "PCF", 3)) {
					// PCF - Protect Configuration command
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						if (c == active_sw_config) {
							if (rxLine.buf[9] == 'U') {
								// Unprotect configuration
								configSw.protection = 0;
							} else if (rxLine.buf[9] == 'P') {
								// Protect configuration
								configSw.protection = 1;
							}
							// save the active configuration
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							if (rxLine.buf[9] == 'U') {
								// Unprotect configuration
								cs.protection = 0;
							} else if (rxLine.buf[9] == 'P') {
								// Protect configuration
								cs.protection = 1;
							}
							// save this non-active configuration
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							// Note: saving takes 120ms!! during which incoming messages are not processed
							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "RST", 3)) {
					// RST command - reset
					goto rst;
/*
					// reset the NAND flash
					retval = nandFlash_reset();
					// TODO reset the FPGA
					fpga_reset();
					// reset micro
					ResetISR();
*/

				}

				// end of message processing section
				rxLine.rptr = rxLine.wptr; // release rx message


			}
		}
	}
	// end of main loop ------------------------------------------------------------------------


	return (0);
}

/*** EOF ***/
